*************************************************************************************
*														*
*	Flame effect simulation for the EASy68k simulator	2006/11/15			*
*														*
*	This code uses double buffering to render a patterm similar to that made	*
*	by flames. It is modified from my rain effect and can be mind numbingly		*
*	slow on anything less than a 1.5GHz PC but it makes some pretty patterns	*
*	and shows how such effects can be made.							*
*														*
*	Four available palettes can be cycled through using the [P] key.			*
*														*
*	This version for Sim68K 3.2 or later							*
*														*
*	More 68000 and other projects can be found on my website at ..			*
*														*
*	 http://www.themotionstore.com/leeedavison/index.html					*
*														*
*	.. or ..												*
*														*
*	 http://members.lycos.co.uk/leeedavison/index.html					*
*														*
*	mail : leeedavison@lycos.co.uk								*
*														*
*************************************************************************************

size	EQU	$40					* both x and y size of the pixel array
size2	EQU	size*size-1				* maximum array index, array is indexed 0 to n

	ORG	$1000

start
	BSR		Initialise			* go setup stuff, pointers, arrays and flags

	MOVE.b	#17,d1			* enable double buffering
	MOVE.b	#92,d0			* set draw mode
	TRAP		#15

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen
	TRAP		#15

* the main loop, after the setup is complete everything is called from within this code

loop
	MOVE.b	#94,d0			* copy screen buffer to main
	TRAP		#15

	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	MOVE.l	d1,-(sp)			* push time on the stack, erk! needed d7

	MOVE.b	#11,d0			* position cursor
	MOVE.w	#$FF00,d1			* clear screen buffer
	TRAP		#15

	BSR.s		c_controls			* go check the [P] key
	BSR.s		gen_array			* generate and draw the new screen data

* now wait for a while. if you have a slow machine this shouldn't slow it down much but
* if you have a fast machine them this will stop the effect running too fast

	MOVE.l	(sp)+,d7			* restore time in 1/100 ths from stack
wait_100ms
	MOVEQ		#8,d0				* get time in 1/100 ths seconds
	TRAP		#15

	SUB.l		d7,d1				* subtract previous time from current time
	CMP.b		#$09,d1			* compare with 9/100ths
	BMI.s		wait_100ms			* loop if time not up yet

* if you want to see how fast it can go comment out the BMI.s above. on my 1.8GHz Celeron
* it does each frame in about 140 to 180ms

	BRA.s		loop				* loop forever

*************************************************************************************
*
* check the [P] key. if [P] has just been pressed step the palette 'page' on to the
* next one

c_controls
	MOVE.l	#'PP',d1			* [P]alette key
	MOVE.b	#19,d0			* check for keypress
	TRAP		#15

	MOVE.w	d1,d0				* copy result, $0000 if 'P' key up, $FFFF if
							* 'P' key down
	EOR.w		d1,lastpal(a3)		* compare with last P key state
	BEQ.s		notogglepal			* if state not changed skip toggle

	AND.w		d0,d1				* make sure just pressed, $FFFF if pressed,
							* $0000 if released
	AND.b		#$40,d1			* mask for palette page offset
	ADD.b		d1,palette(a3)		* toggle palette page if pressed
notogglepal
	MOVE.w	d0,lastpal(a3)		* save this result as last
	RTS

*************************************************************************************
*
* generate the next stage and display it
*
* first modify the 'ground level' pixel intensity. first two 32 bit random numbers
* are made and each bit of these numbers is used to modify the pixel intensity of
* each of the sixty four 'ground level' row of pixels

gen_array
	MOVEQ		#$02,d3			* set spark delta value

	BSR		LAB_RND			* call the PRNG code
	MOVE.l	PRNlword(a3),d0		* get first PRNG result

	BSR		LAB_RND			* call the PRNG code again
	MOVE.l	PRNlword(a3),d1		* get second PRNG result

	MOVEQ		#31,d2			* 32 bits to do
sparkloop
	ADD.l		d0,d0				* shift top bit of first PRNG result into Cb
	BCC.s		nospark1			* skip set if not set

	ADD.b		d3,(a1,d2)			* add delta to spark
nospark1
	ADD.l		d1,d1				* shift top bit of second PRNG result into Cb
	BCC.s		nospark2			* skip set if not set

	ADD.b		d3,32(a1,d2)		* add delta to spark
nospark2
	DBF		d2,sparkloop		* loop if more to do

* for each pixel the new pixel value = (sum of the four surrounding pixels)\4 plus a
* modifier. while this could be computed it is quicker, and easier, to use the sum as
* an index to a pre-computed lookup table

	MOVEQ		#size-1,d7			* set initial y index
	MOVEQ		#$1C,d2			* set initial screen y1 co-ordinate
	MOVEQ		#$1F,d4			* set initial screen y2 co-ordinate, make a
							* small square

	MOVEQ		#size-1,d5			* set initial x-1 index, time is saved by
							* moving the initial setting of this outside
							* both loops. within the loops it stays in range
							* by being masked at each itteration

	LEA		size*size(a2),a0		* initial y base address, array 2
	LEA		size*size(a1),a5		* initial y base address, array 1

* lots of time is saved by making use of the registers. for each y loop the next y-1, y
* and y+1 base addresses are calculated for array 1 as well as the next y base address for
* array 2

gen_yloop
	LEA		-size(a0),a0		* next y base address, array 2
	LEA		-size(a5),a5		* next y base address, array 1
	LEA		-size(a5),a4		* next y-1 base address, array 1
	TST.w		d7				* y index
	BNE.s		gen_yloop_e			* if not zeroeth element skip address reset

* for the 'ground level' row of pixels there is no lower row so the sum is made from
* twice the pixel's own intensity plus that of it's horizontal neighbours. this is done
* by adjusting the y - 1 pointer for what would have been the row below to point to this
* row instead

	MOVEA.l	a5,a4				* else set y - 1 to bottom

* calculate the new screen y co-ordinate for this row of pixels. time is saved by having
* this outside the x loop

gen_yloop_e
	ADDQ.w	#4,d2				* calculate the new screen y1 co-ordinate
	ADDQ.w	#4,d4				* calculate the new screen y2 co-ordinate

* set the screen x2 co-ordinate for the first pixel on this row, this saves time as the
* calculation for the next screen x co-ordinates reduces to a copy, an add and a subtract

	MOVEQ		#size-1,d6			* set initial x index
	MOVEQ		#$1F,d3			* set initial screen x2 co-ordinate

* we have y-1 and y in a4 and a5, now for each x loop first calculate the x-1 and x+1
* values and use these as offsets to the y base addresses to access the values of the
* three adjoining pixels to the current pixel

* the x loop is repeated once for every pixel so any cycles that can be saved from inside
* this loop will save size*size cycles per frame which for a 64 x 64 array is 4096 cycles

gen_xloop
	MOVEQ		#size-1,d0			* set mask and clear the three highest bytes

	SUBQ.w	#1,d5				* decrement x - 1
	AND.w		d0,d5				* mask to array size

	MOVE.w	d6,d1				* set x index
	ADDQ.w	#1,d1				* add 1
	AND.w		d0,d1				* mask to array size

* here we have x-1, x and x+1 in d5, d6 and d1. using these with a4 and a5 allows easy
* access to the pixel elements

* now get the total for the pixel at x,y and the three below it. this value can sometimes
* cause a carry on the 'ground level' row of pixels but this adds to the effect as it
* gives an occasional gout of no flame in the hottest parts of the flame base

	MOVE.b	(a5,d6.w),d0		* get pixel intensity ( 0,  0)

	ADD.b		(a4,d1.w),d0		* add pixel intensity (+1, -1)
	ADD.b		(a4,d5.w),d0		* add pixel intensity (-1, -1)
	ADD.b		(a4,d6.w),d0		* add pixel intensity ( 0, -1)

	MOVE.b	(a6,d0.w),d0		* get pre-computed new pixel intensity
	MOVE.b	d0,(a0,d6.w)		* save new pixel intensity

* select between the four palette effect 'pages'. OR is used so the total of page and
* intensity never overflows the byte value

	OR.b		palette(a3),d0		* OR with palette flag, select page

* now set the intensity for this pixel as the draw colour and the fill colour

	ADD.w		d0,d0				* * 2, high byte cleared by previous MOVEQ d0
	ADD.w		d0,d0				* * 4 bytes per longword colour value
	MOVE.l	(a3,d0.w),d1		* get the longword colour value from the table

	MOVEQ		#80,d0			* set pen colour
	TRAP		#15

	MOVEQ		#81,d0			* set fill colour
	TRAP		#15

* calculate the screen x co-ordinate for this pixel

	ADDQ.w	#4,d3				* calculate the new screen x2 co-ordinate
	MOVE.w	d3,d1				* copy to the screen x1 co-ordinate
	SUBQ.w	#3,d1				* make a small square

* now draw a small 4 x 4 square of pixels

	MOVEQ		#87,d0			* draw filled rectangle
	TRAP		#15				* do draw

	DBF		d6,gen_xloop		* decrement x and loop for next

	DBF		d7,gen_yloop		* decrement y and loop for next

	EXG		a1,a2				* swap the current and last arrays
	RTS

*************************************************************************************
*
* setup stuff. set the pointers, initialise the PRNG seed, clear the palette page
* and key flag and clear the arrays

Initialise
	LEA		(LAB_VARS,PC),a3		* get pointer to variables
	LEA		(flametable,PC),a6	* get pointer to flame table
	LEA		array1(a3),a1		* get pointer to array 1
	LEA		array2(a3),a2		* get pointer to array 2
	MOVE.l	#$31415927,PRNlword(a3)	* set PRNG seed. this can be any value
							* except zero, changing this will produce
							* a different start point for the effect
	MOVEQ		#0,d0				* clear d0
	MOVE.w	d0,palette(a3)		* clear the palette page
	MOVE.w	d0,lastpal(a3)		* clear the last P key flag
	MOVE.l	#size*size/2-1,d1		* get array size/2
	MOVEA.l	a1,a4				* copy array memory start
clear_loop
	MOVE.l	d0,(a4)+			* clear longword
	DBF		d1,clear_loop		* decrement count and loop for next

	RTS

*************************************************************************************
*
* this is the code that generates the pseudo random sequence. a seed word located in
* PRNlword(a3) is loaded into a register before being operated on to generate the
* next number in the sequence. this number is then saved as the seed for the next
* time it's called.
*
* this code is adapted from the 32 bit version of RND(n) used in EhBASIC68. Taking
* the 33rd next number is slower but means that there is no bit overlap between
* successive returned values as all 32 bits of the value are used.

LAB_RND
	MOVEM.l	d0-d2,-(sp)			* save d0, d1 and d2
	MOVE.l	PRNlword(a3),d0		* get current seed longword
	MOVEQ		#$AF-$100,d1		* set EOR value
	MOVEQ		#20,d2			* do this 33 times
Ninc0
	ADD.l		d0,d0				* shift left 1 bit
	BCC.s		Ninc1				* if bit not set skip feedback

	EOR.b		d1,d0				* do Galois LFSR feedback
Ninc1
	DBF		d2,Ninc0			* loop

	MOVE.l	d0,PRNlword(a3)		* save back to seed longword
	MOVEM.l	(sp)+,d0-d2			* restore d0, d1 and d2

	RTS

*************************************************************************************
*
* the flame table. the next value of a flame pixel is made from the sum of the values
* of that pixel and the three below it divided by four, this is further modified that
* if as a result of the divide the carry is set then the value is incremented, but if
* the carry was clear the value is decremented twice. also the value is checked and
* kept within the bounds of $00 to $3F
*
* while this can be computed it is far easier, and quicker, to just make a table of
* pre computed results and use the pixel sum as an index to that table, this also
* allows other pre-computed effects to be added without increasing processing time

flametable
	dc.b	$00,$00,$00,$00,$00,$00,$02,$02,$00,$00,$03,$03,$01,$01,$04,$04
	dc.b	$02,$02,$05,$05,$03,$03,$06,$06,$04,$04,$07,$07,$05,$05,$08,$08
	dc.b	$06,$06,$09,$09,$07,$07,$0A,$0A,$08,$08,$0B,$0B,$09,$09,$0C,$0C
	dc.b	$0A,$0A,$0D,$0D,$0B,$0B,$0E,$0E,$0C,$0C,$0F,$0F,$0D,$0D,$10,$10
	dc.b	$0E,$0E,$11,$11,$0F,$0F,$12,$12,$10,$10,$13,$13,$11,$11,$14,$14
	dc.b	$12,$12,$15,$15,$13,$13,$16,$16,$14,$14,$17,$17,$15,$15,$18,$18
	dc.b	$16,$16,$19,$19,$17,$17,$1A,$1A,$18,$18,$1B,$1B,$19,$19,$1C,$1C
	dc.b	$1A,$1A,$1D,$1D,$1B,$1B,$1E,$1E,$1C,$1C,$1F,$1F,$1D,$1D,$20,$20
	dc.b	$1E,$1E,$21,$21,$1F,$1F,$22,$22,$20,$20,$23,$23,$21,$21,$24,$24
	dc.b	$22,$22,$25,$25,$23,$23,$26,$26,$24,$24,$27,$27,$25,$25,$28,$28
	dc.b	$26,$26,$29,$29,$27,$27,$2A,$2A,$28,$28,$2B,$2B,$29,$29,$2C,$2C
	dc.b	$2A,$2A,$2D,$2D,$2B,$2B,$2E,$2E,$2C,$2C,$2F,$2F,$2D,$2D,$30,$30
	dc.b	$2E,$2E,$31,$31,$2F,$2F,$32,$32,$30,$30,$33,$33,$31,$31,$34,$34
	dc.b	$32,$32,$35,$35,$33,$33,$36,$36,$34,$34,$37,$37,$35,$35,$38,$38
	dc.b	$36,$36,$39,$39,$37,$37,$3A,$3A,$38,$38,$3B,$3B,$39,$39,$3C,$3C
	dc.b	$3A,$3A,$3D,$3D,$3B,$3B,$3E,$3E,$3C,$3C,$3F,$3F,$3D,$3D,$3F,$3F

*************************************************************************************
*
* variables used

LAB_VARS						* points to variables base

* colour table for converting byte colour values into longword colour values, this saves
* 16 cycles for each conversion or 65536 cycles per 64 * 64 frame

* another advantage of using a table is that a colour map may be applied that transforms
* the linear grey scale to non linear or even to a coloured scale for such things as
* flame or inverse effects. colours are $00bbggrr

* flame palette
	dc.l	$000000,$000005,$00000B,$000012,$00001A,$000022,$00002B,$000034
	dc.l	$00003F,$000049,$000054,$00005F,$00006A,$000076,$000082,$00008D
	dc.l	$000098,$0000A4,$0000AE,$0000B9,$0000C3,$0000CD,$0000D7,$0000DF
	dc.l	$0000E7,$0000EF,$0000F5,$0000FB,$0002FD,$0007FD,$000EFD,$0015FD
	dc.l	$001DFD,$0026FD,$0030FD,$003AFD,$0044FD,$0050FD,$005AFD,$0066FD
	dc.l	$0071FD,$007DFD,$0088FD,$0093FD,$009EFD,$00A8FD,$00B2FD,$00BBFD
	dc.l	$00C4FD,$00CBFD,$00D2FD,$00D7FD,$00DCFD,$00DFFD,$00E2FD,$03E0FD
	dc.l	$10E4FD,$20E8FD,$30ECFD,$40F0FD,$50F4FE,$60F8FE,$70FCFE,$7FFFFF

* normal palette
	dc.l	$000000,$010101,$020202,$030303,$040404,$050505,$060606,$070707
	dc.l	$080808,$090909,$0A0A0A,$0B0B0B,$0C0C0C,$0D0D0D,$0E0E0E,$0F0F0F
	dc.l	$101010,$111111,$121212,$131313,$141414,$151515,$161616,$171717
	dc.l	$181818,$191919,$1A1A1A,$1B1B1B,$1C1C1C,$1D1D1D,$1E1E1E,$1F1F1F
	dc.l	$202020,$212121,$222222,$232323,$242424,$252525,$262626,$272727
	dc.l	$282828,$292929,$2A2A2A,$2B2B2B,$2C2C2C,$2D2D2D,$2E2E2E,$2F2F2F
	dc.l	$303030,$313131,$323232,$333333,$343434,$353535,$363636,$373737
	dc.l	$383838,$393939,$3A3A3A,$3B3B3B,$3C3C3C,$3D3D3D,$3E3E3E,$3F3F3F

* 4 x scale palette
	dc.l	$000000,$040404,$080808,$0C0C0C,$101010,$141414,$181818,$1C1C1C
	dc.l	$202020,$242424,$282828,$2C2C2C,$303030,$343434,$383838,$3C3C3C
	dc.l	$404040,$444444,$484848,$4C4C4C,$505050,$545454,$585858,$5C5C5C
	dc.l	$606060,$646464,$686868,$6C6C6C,$707070,$747474,$787878,$7C7C7C
	dc.l	$808080,$848484,$888888,$8C8C8C,$909090,$949494,$989898,$9C9C9C
	dc.l	$A0A0A0,$A4A4A4,$A8A8A8,$ACACAC,$B0B0B0,$B4B4B4,$B8B8B8,$BCBCBC
	dc.l	$C0C0C0,$C4C4C4,$C8C8C8,$CCCCCC,$D0D0D0,$D4D4D4,$D8D8D8,$DCDCDC
	dc.l	$E0E0E0,$E4E4E4,$E8E8E8,$ECECEC,$F0F0F0,$F4F4F4,$F8F8F8,$FFFFFF

* 4 x scale inverse palette
	dc.l	$FFFFFF,$F8F8F8,$F4F4F4,$F0F0F0,$ECECEC,$E8E8E8,$E4E4E4,$E0E0E0
	dc.l	$DCDCDC,$D8D8D8,$D4D4D4,$D0D0D0,$CCCCCC,$C8C8C8,$C4C4C4,$C0C0C0
	dc.l	$BCBCBC,$B8B8B8,$B4B4B4,$B0B0B0,$ACACAC,$A8A8A8,$A4A4A4,$A0A0A0
	dc.l	$9C9C9C,$989898,$949494,$909090,$8C8C8C,$888888,$848484,$808080
	dc.l	$7C7C7C,$787878,$747474,$707070,$6C6C6C,$686868,$646464,$606060
	dc.l	$5C5C5C,$585858,$545454,$505050,$4C4C4C,$484848,$444444,$404040
	dc.l	$3C3C3C,$383838,$343434,$303030,$2C2C2C,$282828,$242424,$202020
	dc.l	$1C1C1C,$181818,$141414,$101010,$0C0C0C,$080808,$040404,$000000

colourend

	OFFSET	colourend-LAB_VARS	* use relative addressing

PRNlword	ds.l	1				* PRNG seed long word

palette	ds.w	1				* palette page
lastpal	ds.w	1				* last P key flag

array1	ds.b	size*size			* make space for the first array
array2	ds.b	size*size			* make space for the second array
arrayend

	ORG	LAB_VARS+arrayend			* use program counter addressing

	END	start
